.wy-table-responsive .wrap-table td,
.wy-table-responsive .wrap-table th {
    white-space: normal !important;
    word-wrap: break-word;
    overflow-wrap: break-word;
}

.wrap-table {
    table-layout: fixed;
    width: 100%;
}

.wrap-table th, .wrap-table td {
    word-wrap: break-word;
    overflow-wrap: break-word;
}
