#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Library of generic processors.

Subpackage containing a library of general purpose processors that the user can either use directly or overload
them to implement their analytical tasks.
"""
